from http import HTTPStatus

from django.conf import settings as django_settings
from django.contrib.auth import REDIRECT_FIELD_NAME, authenticate, login
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import resolve_url
from django.utils.http import url_has_allowed_host_and_scheme  # private API
from django.views.generic import View

from . import settings

try:
    from django.contrib.auth.views import RedirectURLMixin  # private API
except ImportError:  # Django < 4.1
    # Added to Django in https://github.com/django/django/pull/15608
    class RedirectURLMixin:
        next_page = None
        redirect_field_name = REDIRECT_FIELD_NAME
        success_url_allowed_hosts = set()

        def get_success_url(self):
            return self.get_redirect_url() or self.get_default_redirect_url()

        def get_redirect_url(self):
            """Return the user-originating redirect URL if it's safe."""
            redirect_to = self.request.POST.get(
                self.redirect_field_name, self.request.GET.get(self.redirect_field_name)
            )
            url_is_safe = url_has_allowed_host_and_scheme(
                url=redirect_to,
                allowed_hosts=self.get_success_url_allowed_hosts(),
                require_https=self.request.is_secure(),
            )
            return redirect_to if url_is_safe else ""

        def get_success_url_allowed_hosts(self):
            return {self.request.get_host(), *self.success_url_allowed_hosts}

        def get_default_redirect_url(self):
            """Return the default redirect URL."""
            if self.next_page:
                return resolve_url(self.next_page)
            else:  # pragma: no cover
                raise ImproperlyConfigured("no URL to redirect to; provide a next_page")


__all__ = ["LoginView"]


class LoginView(RedirectURLMixin, View):
    """
    Look for a signed token in the URL of a GET request and log a user in.

    If a valid token is found, the user is redirected to the URL specified in
    the ``next`` query string parameter or the ``next_page`` attribute of the
    view. ``next_page`` defaults to :setting:`LOGIN_REDIRECT_URL`.

    If a ``scope`` attribute is set, a :ref:`scoped token <Scoped tokens>` is
    expected.

    If a ``max_age`` attribute is set, override the :data:`SESAME_MAX_AGE`
    setting.

    In addition to ``next_page``, :class:`LoginView` also supports
    ``redirect_field_name``, ``success_url_allowed_hosts``, and
    ``get_default_redirect_url()``. These APIs behave like their counterparts
    in Django's built-in :class:`~django.contrib.auth.views.LoginView`.

    """

    scope = ""
    max_age = None
    next_page = django_settings.LOGIN_REDIRECT_URL

    def get(self, request):
        if not hasattr(request, "user"):
            raise ImproperlyConfigured("LoginView requires django.contrib.auth")

        sesame = request.GET.get(settings.TOKEN_NAME)
        if sesame is None:
            return self.login_failed()

        user = authenticate(
            request,
            sesame=sesame,
            scope=self.scope,
            max_age=self.max_age,
        )
        if user is None:
            return self.login_failed()

        login(request, user)  # updates the last login date

        return self.login_success()

    def login_failed(self):
        raise PermissionDenied

    def login_success(self):
        if self.next_page is None:
            return HttpResponse(status=HTTPStatus.NO_CONTENT)
        else:
            return HttpResponseRedirect(self.get_success_url())



import struct
import uuid

from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string

from . import settings

__all__ = [
    "BasePacker",
    "ShortPacker",
    "UnsignedShortPacker",
    "LongPacker",
    "UnsignedLongPacker",
    "LongLongPacker",
    "UnsignedLongLongPacker",
    "UUIDPacker",
    "BytesPacker",
    "StrPacker",
    "packer",
]


class BasePacker:
    """
    Abstract base class for packers.

    """

    def pack_pk(self, user_pk):
        """
        Create a short representation of the primary key of a user.

        Return :class:`bytes`.

        """

    def unpack_pk(self, data):
        """
        Extract the primary key of a user from a signed token.

        ``data`` contains :class:`bytes`.

        Return the primary key and the remaining data as :class:`bytes`.

        """


class StructPackerMeta(type):
    def __new__(cls, name, bases, namespace, **kwds):
        if "size" not in namespace and "fmt" in namespace:
            namespace["size"] = struct.calcsize(namespace["fmt"])
        return super().__new__(cls, name, bases, namespace, **kwds)


class StructPacker(BasePacker, metaclass=StructPackerMeta):
    fmt = ""

    @classmethod
    def pack_pk(cls, user_pk):
        return struct.pack(cls.fmt, user_pk)

    @classmethod
    def unpack_pk(cls, data):
        (user_pk,) = struct.unpack(cls.fmt, data[: cls.size])
        return user_pk, data[cls.size :]


class ShortPacker(StructPacker):
    fmt = "!h"


class UnsignedShortPacker(StructPacker):
    fmt = "!H"


class LongPacker(StructPacker):
    fmt = "!l"


IntPacker = LongPacker  # for backwards-compatibility


class UnsignedLongPacker(StructPacker):
    fmt = "!L"


UnsignedIntPacker = UnsignedLongPacker  # for consistency


class LongLongPacker(StructPacker):
    fmt = "!q"


class UnsignedLongLongPacker(StructPacker):
    fmt = "!Q"


class UUIDPacker(BasePacker):
    @staticmethod
    def pack_pk(user_pk):
        return user_pk.bytes

    @staticmethod
    def unpack_pk(data):
        return uuid.UUID(bytes=data[:16]), data[16:]


class BytesPacker(BasePacker):
    """
    Generic packer for bytestrings, from 0 to 255 bytes.

    In many cases, primary keys stored as bytes are likely to be fixed-length,
    which doesn't require a variable length encoding scheme.

    """

    @staticmethod
    def pack_pk(user_pk):
        length = len(user_pk)
        if length > 255:
            raise ValueError("primary key is too large (%d bytes)" % length)
        return bytes([length]) + user_pk

    @staticmethod
    def unpack_pk(data):
        length = data[0]
        return data[1 : length + 1], data[length + 1 :]


class StrPacker(BytesPacker):
    """
    Generic packer for strings, from 0 to 255 UTF-8 encoded bytes.

    """

    @staticmethod
    def pack_pk(user_pk):
        user_pk = user_pk.encode()
        length = len(user_pk)
        if length > 255:
            raise ValueError("primary key is too large (%d UTF-8 bytes)" % length)
        return bytes([length]) + user_pk

    @staticmethod
    def unpack_pk(data):
        length = data[0]
        return data[1 : length + 1].decode(), data[length + 1 :]


PACKERS = {
    # 2 bytes
    "SmallAutoField": ShortPacker,
    "SmallIntegerField": ShortPacker,
    "PositiveSmallIntegerField": UnsignedShortPacker,
    # 4 bytes
    "AutoField": LongPacker,
    "IntegerField": LongPacker,
    "PositiveIntegerField": UnsignedLongPacker,
    # 8 bytes
    "BigAutoField": LongLongPacker,
    "BigIntegerField": LongLongPacker,
    "PositiveBigIntegerField": UnsignedLongLongPacker,
    # 16 bytes
    "UUIDField": UUIDPacker,
    # Variable length
    "BinaryField": BytesPacker,
    "CharField": StrPacker,
    "TextField": StrPacker,
}


def get_packer():
    if settings.PACKER is None:
        User = get_user_model()
        if settings.PRIMARY_KEY_FIELD == "pk":
            pk_field = User._meta.pk
        else:
            pk_field = User._meta.get_field(settings.PRIMARY_KEY_FIELD)
            if not pk_field.unique:
                raise ImproperlyConfigured(
                    f"{User._meta.label}.{settings.PRIMARY_KEY_FIELD} isn't unique"
                )
        pk_type = pk_field.get_internal_type()
        try:
            Packer = PACKERS[pk_type]
        except KeyError:
            raise NotImplementedError(pk_type + " primary keys aren't supported")
    else:
        Packer = import_string(settings.PACKER)
    return Packer()


packer = get_packer()



from django.contrib.auth import backends as auth_backends
from django.contrib.auth import get_user_model

from . import settings
from .tokens import parse_token

__all__ = ["ModelBackend", "SesameBackendMixin"]


class SesameBackendMixin:
    """
    Mix this class in an authentication backend providing ``get_user(user_id)``
    to create an authentication backend usable with django-sesame.

    """

    def authenticate(self, request, sesame, scope="", max_age=None):
        """
        If ``sesame`` is a valid token, return the user. Else, return :obj:`None`.

        If ``scope`` is set, a :ref:`scoped token <Scoped tokens>` is expected.

        If ``max_age`` is set, override the :data:`SESAME_MAX_AGE` setting.

        ``request`` is an :class:`~django.http.HttpRequest` or :obj:`None`.

        """
        # This check shouldn't be necessary, but it can avoid problems like
        # issue #37 and Django's built-in backends include similar checks.
        if sesame is None:
            return None
        return parse_token(sesame, self.get_user, scope, max_age)


class ModelBackend(SesameBackendMixin, auth_backends.ModelBackend):
    """
    Authentication backend that authenticates users with django-sesame tokens.

    It inherits :class:`SesameBackendMixin` and Django's built-in
    :class:`~django.contrib.auth.backends.ModelBackend`.

    Extending the default value of the :setting:`AUTHENTICATION_BACKENDS` setting
    to add  ``"sesame.backends.ModelBackend"`` looks like:

    .. code-block:: python

        AUTHENTICATION_BACKENDS = [
            "django.contrib.auth.backends.ModelBackend",
            "sesame.backends.ModelBackend",
        ]

    """

    def get_user(self, user_id):
        """
        Fetch user from the database by primary key.

        The field used by django-sesame as a primary key can be configured with
        the :data:`SESAME_PRIMARY_KEY_FIELD` setting.

        Return :obj:`None` if no active user is found.

        """
        User = get_user_model()
        try:
            user = User._default_manager.get(**{settings.PRIMARY_KEY_FIELD: user_id})
        except User.DoesNotExist:
            return None
        if self.user_can_authenticate(user):
            return user
        else:
            return None



import datetime
import hashlib
import importlib
import sys

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

DEFAULTS = {
    # Generating URLs
    "TOKEN_NAME": "sesame",
    # Tokens lifecycle
    "MAX_AGE": None,
    "ONE_TIME": False,
    "INVALIDATE_ON_PASSWORD_CHANGE": True,
    # Custom primary keys
    "PACKER": None,
    "PRIMARY_KEY_FIELD": "pk",
    # Tokens
    "TOKENS": ["sesame.tokens_v2", "sesame.tokens_v1"],
    # Tokens v2
    "KEY": "",
    # We want a short signature in order to keep tokens short. A 10-bytes
    # signature has about 1.2e24 possible values, which is sufficient here.
    "SIGNATURE_SIZE": 10,
    # Tokens v1
    "SALT": "sesame",
    # These parameters aren't updated anymore. Tokens v2 are recommended.
    "DIGEST": hashlib.md5,
    "ITERATIONS": 10000,
}

__all__ = list(DEFAULTS)


def derive_key(secret_key, key):
    """
    Make a 64-bytes key from Django's ``secret_key`` and django-sesame's ``key``.

    Include settings in the key to invalidate tokens when these settings change.
    This ensures that tokens generated with one packer cannot be misinterpreted
    by another packer, for example.

    """
    global MAX_AGE, PACKER, PRIMARY_KEY_FIELD
    return hashlib.blake2b(
        "|".join(
            [
                # Usually a str but Django also supports bytes.
                str(secret_key),
                # Treat key like secret_key for consistency.
                str(key),
                # Changing MAX_AGE is allowed as long as it is not None.
                "max_age" if MAX_AGE is not None else "",
                PACKER if PACKER is not None else "",
                PRIMARY_KEY_FIELD,
            ]
        ).encode(),
        person=b"sesame.settings",
    ).digest()


# load() also works for reloading settings, which is useful for testing.


def load():
    module = sys.modules[__name__]
    for name, default in DEFAULTS.items():
        setattr(module, name, getattr(settings, "SESAME_" + name, default))

    global KEY, MAX_AGE, SIGNING_KEY, TOKENS, VERIFICATION_KEYS

    # Support defining MAX_AGE as a timedelta rather than a number of seconds.
    if isinstance(MAX_AGE, datetime.timedelta):
        MAX_AGE = MAX_AGE.total_seconds()

    # Import token creation and parsing modules.
    TOKENS = [importlib.import_module(tokens) for tokens in TOKENS]

    # Derive signing and verification keys.
    SIGNING_KEY = derive_key(settings.SECRET_KEY, KEY)
    VERIFICATION_KEYS = [SIGNING_KEY] + [
        derive_key(secret_key, KEY)
        for secret_key in getattr(settings, "SECRET_KEY_FALLBACKS", [])
    ]


load()


# Django's checks framework was designed to run such checks. Unfortunately,
# there's no way to guarantee that a check would be discovered, because
# django-sesame never required adding the "sesame" app to INSTALLED_APPS.

# The benefits of writing this check with the checks framework don't justify
# adding another step to the installation instructions. Raising an exception
# is good enough.


def check():
    global MAX_AGE, INVALIDATE_ON_PASSWORD_CHANGE
    if MAX_AGE is None and not INVALIDATE_ON_PASSWORD_CHANGE:
        raise ImproperlyConfigured(
            "insecure configuration: set SESAME_MAX_AGE to a low value "
            "or set SESAME_INVALIDATE_ON_PASSWORD_CHANGE to True"
        )


check()



import logging
import re

import django
from django.core import signing
from django.utils import crypto

from . import packers, settings

__all__ = ["create_token", "detect_token", "parse_token"]

logger = logging.getLogger("sesame")


def get_revocation_key(user):
    """
    When the value returned by this method changes, this revocates tokens.

    It is derived from the hashed password so that changing the password
    revokes tokens.

    For one-time tokens, it also contains the last login datetime so that
    logging in revokes existing tokens.

    """
    data = ""
    if settings.INVALIDATE_ON_PASSWORD_CHANGE:
        data += user.password
    if settings.ONE_TIME:
        data += str(user.last_login)
    # The password is expected to be a secure hash but we hash it again
    # for additional safety. We default to MD5 to minimize the length of
    # the token. (Remember, if an attacker obtains the URL, he can already
    # log in. This isn't high security.)
    return crypto.pbkdf2(
        data,
        settings.SALT,
        settings.ITERATIONS,
        digest=settings.DIGEST,
    )


def get_signer():
    kwargs = {"salt": settings.SALT}
    if django.VERSION[:2] >= (3, 1):  # pragma: no cover
        kwargs["algorithm"] = "sha1"
    if settings.MAX_AGE is None:
        return signing.Signer(**kwargs)
    else:
        return signing.TimestampSigner(**kwargs)


signer = get_signer()


def sign(data):
    """
    Create a URL-safe, signed token from ``data``.

    """
    data = signing.b64_encode(data).decode()
    return signer.sign(data)


def unsign(token):
    """
    Extract the data from a signed ``token``.

    """
    if settings.MAX_AGE is None:
        data = signer.unsign(token)
    else:
        data = signer.unsign(token, max_age=settings.MAX_AGE)
    return signing.b64_decode(data.encode())


def create_token(user, scope=""):
    """
    Create a v1 signed token for a user.

    """
    if scope != "":
        raise NotImplementedError("v1 tokens don't support scope")
    primary_key = packers.packer.pack_pk(getattr(user, settings.PRIMARY_KEY_FIELD))
    key = get_revocation_key(user)
    return sign(primary_key + key)


def parse_token(token, get_user, scope="", max_age=None):
    """
    Obtain a user from a v1 signed token.

    """
    if scope != "":
        raise NotImplementedError("v1 tokens don't support scope")
    if max_age is not None:
        raise NotImplementedError("v1 tokens don't support max_age")

    try:
        data = unsign(token)
    except signing.SignatureExpired:
        logger.debug("Expired token: %s", token)
        return None
    except signing.BadSignature:
        logger.debug("Bad token: %s", token)
        return None
    except Exception:
        logger.exception(
            "Valid signature but unexpected token; if you enabled "
            "or disabled SESAME_MAX_AGE, you must regenerate tokens"
        )
        return None

    try:
        user_pk, key = packers.packer.unpack_pk(data)
    except Exception:
        logger.exception(
            "Valid signature but unexpected token; if you changed "
            "SESAME_PACKER, you must regenerate tokens"
        )
        return None

    user = get_user(user_pk)
    if user is None:
        logger.debug("Unknown or inactive user: %s", user_pk)
        return None
    if not crypto.constant_time_compare(key, get_revocation_key(user)):
        logger.debug("Invalid token: %s", token)
        return None
    logger.debug("Valid token for user %s: %s", user, token)

    return user


def get_token_re():
    if settings.MAX_AGE is None:
        # Size of primary key and revocation key depends on SESAME_PACKER and
        # SESAME_DIGEST. Default is 4 + 16 = 20 bytes = 27 Base64 characters.
        # Minimum "sensible" size is 1 + 2 = 3 bytes = 4 Base64 characters.
        return re.compile(r"[A-Za-z0-9-_]{4,}:[A-Za-z0-9-_]{27}")

    else:
        # All timestamps use 6 Base62 characters because 100000 in Base62 is
        # 1999-01-12T09:20:32Z, before django-sesame existed.
        return re.compile(r"[A-Za-z0-9-_]{4,}:[0-9A-Za-z]{6}:[A-Za-z0-9-_]{27}")


token_re = get_token_re()


def detect_token(token):
    """
    Tell whether token may be a v1 signed token.

    """
    return token_re.fullmatch(token) is not None



import base64
import datetime
import hashlib
import hmac
import logging
import re
import struct
import time

from . import packers, settings

__all__ = ["create_token", "detect_token", "parse_token"]

logger = logging.getLogger("sesame")

TIMESTAMP_OFFSET = 1577836800  # 2020-01-01T00:00:00Z


def pack_timestamp():
    """
    When SESAME_MAX_AGE is enabled, encode the time in seconds since the epoch.

    Return bytes.

    """
    if settings.MAX_AGE is None:
        return b""
    timestamp = int(time.time()) - TIMESTAMP_OFFSET
    return struct.pack("!i", timestamp)


def unpack_timestamp(data):
    """
    When SESAME_MAX_AGE is enabled, extract the timestamp and calculate the age.

    Return an age in seconds or None and the remaining bytes.

    """
    if settings.MAX_AGE is None:
        return None, data
    # If data contains less than 4 bytes, this raises struct.error.
    (timestamp,), data = struct.unpack("!i", data[:4]), data[4:]
    return int(time.time()) - TIMESTAMP_OFFSET - timestamp, data


HASH_SIZES = {
    "pbkdf2_sha256": 44,
    "pbkdf2_sha1": 28,
    "argon2": 22,  # in Argon2 v1.3; previously 86
    "bcrypt_sha256": 31,  # salt (22) + hash (31)
    "bcrypt": 31,  # salt (22) + hash (31)
    "sha1": 40,  # hex, not base64
    "md5": 32,  # hex, not base64
    "crypt": 11,  # salt (2) + hash (11)
}


def get_revocation_key(user):
    """
    When the value returned by this method changes, this revocates tokens.

    It is derived from the hashed password so that changing the password
    revokes tokens.

    For one-time tokens, it also contains the last login datetime so that
    logging in revokes existing tokens.

    """
    data = ""

    # Tokens generated by django-sesame are more likely to leak than hashed
    # passwords. To minimize the information tokens might be revealing, we'd
    # like to use only hashes, excluding salts, as suggested in issue #40.

    # Since we're hashing the result again with a cryptographic hash function,
    # this isn't supposed to make a difference in practice. But it alleviates
    # concerns about sending data derived from hashed passwords into the wild.

    # Hashed passwords may be in various formats:
    # 1. "[<algorithm>$]?[<parameters>$]*[<salt>$?]?<hash>", if set_password()
    #    was called with a built-in hasher. Unfortunatly, the bcrypt (and
    #    crypt) hashers don't include a "$" between the salt and the hash, so
    #    we can't split on this marker. Instead we hardcode hash lengths.
    # 2. "!<40 random characters>", if set_unusable_password() was called.
    # 3. Anything else, if set_password() was called with a custom hasher or
    #    if a custom authentication backend is used.

    # An alternative would be to rely on user.get_session_auth_hash(), which
    # has the advantage of being a public API. It's a HMAC-SHA256 of the whole
    # password hash. However, it's designed for a slightly different purpose,
    # so I'm not comfortable reusing it. Also, for clarity, I don't want to
    # chain more cryptographic operations than needed.

    if settings.INVALIDATE_ON_PASSWORD_CHANGE and user.password is not None:
        algorithm = user.password.partition("$")[0]
        try:
            hash_size = HASH_SIZES[algorithm]
        except KeyError:
            data += user.password
        else:
            data += user.password[-hash_size:]

    if settings.ONE_TIME and user.last_login is not None:
        data += user.last_login.isoformat()

    return data.encode()


def sign(data, key, size):
    """
    Create a MAC with keyed hashing.

    """
    return hashlib.blake2b(
        data,
        digest_size=size,
        key=key,
        person=b"sesame.tokens_v2",
    ).digest()


def create_token(user, scope=""):
    """
    Create a v2 signed token for a user.

    """
    primary_key = packers.packer.pack_pk(getattr(user, settings.PRIMARY_KEY_FIELD))
    timestamp = pack_timestamp()
    revocation_key = get_revocation_key(user)

    signature = sign(
        primary_key + timestamp + revocation_key + scope.encode(),
        settings.SIGNING_KEY,
        settings.SIGNATURE_SIZE,
    )

    # If the revocation key changes, the signature becomes invalid, so we
    # don't need to include a hash of the revocation key in the token.
    data = primary_key + timestamp + signature
    token = base64.urlsafe_b64encode(data).rstrip(b"=")
    return token.decode()


def parse_token(token, get_user, scope="", max_age=None):
    """
    Obtain a user from a v2 signed token.

    """
    token = token.encode()

    # Below, error messages should give a hint to developers debugging apps
    # but remain sufficiently generic for the common situation where tokens
    # get truncated by accident.

    try:
        data = base64.urlsafe_b64decode(token + b"=" * (-len(token) % 4))
    except Exception:
        logger.debug("Bad token: cannot decode token")
        return None

    # Extract user primary key, token age, and signature from token.

    try:
        user_pk, timestamp_and_signature = packers.packer.unpack_pk(data)
    except Exception:
        logger.debug("Bad token: cannot extract primary key")
        return None

    try:
        age, signature = unpack_timestamp(timestamp_and_signature)
    except Exception:
        logger.debug("Bad token: cannot extract timestamp")
        return None

    if len(signature) != settings.SIGNATURE_SIZE:
        logger.debug("Bad token: cannot extract signature")
        return None

    # Since we don't include the revocation key in the token, we need to fetch
    # the user in the database before we can verify the signature. Usually,
    # it's best to verify the signature before doing anything with a message.

    # An attacker could craft tokens to fetch arbitrary users by primary key,
    # like they can fetch arbitrary users by username on a login form.
    # Determining whether there's a user with a given primary key via a timing
    # attack is acceptable within django-sesame's threat model.

    # Check if token is expired. Perform this check first, because it's fast.

    if max_age is None:
        max_age = settings.MAX_AGE
    elif settings.MAX_AGE is None:
        logger.warning(
            "Ignoring max_age argument; "
            "it isn't supported when SESAME_MAX_AGE = None"
        )
    elif isinstance(max_age, datetime.timedelta):
        max_age = max_age.total_seconds()
    if age is not None and age >= max_age:
        logger.debug("Expired token: age = %d seconds", age)
        return None

    # Check if user exists and can log in.

    user = get_user(user_pk)
    if user is None:
        logger.debug(
            "Unknown or inactive user: %s = %r",
            settings.PRIMARY_KEY_FIELD,
            user_pk,
        )
        return None

    # Check if signature is valid

    primary_key_and_timestamp = data[: -settings.SIGNATURE_SIZE]
    revocation_key = get_revocation_key(user)
    for verification_key in settings.VERIFICATION_KEYS:
        expected_signature = sign(
            primary_key_and_timestamp + revocation_key + scope.encode(),
            verification_key,
            settings.SIGNATURE_SIZE,
        )
        if hmac.compare_digest(signature, expected_signature):
            log_scope = "in default scope" if scope == "" else f"in scope {scope}"
            logger.debug("Valid token for user %s %s", user, log_scope)
            return user

    log_scope = "in default scope" if scope == "" else f"in scope {scope}"
    logger.debug("Invalid token for user %s %s", user, log_scope)
    return None


# Tokens are arbitrary Base64-encoded bytestrings. Their size depends on
# SESAME_PACKER, SESAME_MAX_AGE, and SESAME_SIGNATURE_SIZE. Defaults are:
# - without SESAME_MAX_AGE: 4 + 10 = 14 bytes = 19 Base64 characters.
# - with SESAME_MAX_AGE: 4 + 4 + 10 = 18 bytes = 24 Base64 characters.
# Minimum "sensible" size is 1 + 0 + 2 = 3 bytes = 4 Base64 characters.
token_re = re.compile(r"[A-Za-z0-9-_]{4,}")


def detect_token(token):
    """
    Tell whether token may be a v2 signed token.

    """
    return token_re.fullmatch(token) is not None



from urllib.parse import urlencode

from django.contrib.auth import login
from django.contrib.auth.models import AnonymousUser
from django.shortcuts import redirect

from . import settings
from .utils import get_user

__all__ = ["AuthenticationMiddleware"]


class AuthenticationMiddleware:
    """
    Look for a signed token in the URL of any request and log a user in.

    After login, remove the token from the URL with an HTTP redirect.

    .. admonition:: This functionality requires additional setup for Safari.
        :class: warning

        :class:`AuthenticationMiddleware` requires the optional ``ua`` extra to
        prevent :ref:`issues with Safari <Safari issues>`:

        .. code-block:: console

            $ pip install 'django-sesame[ua]'

    In the :setting:`MIDDLEWARE` setting,
    ``"sesame.middleware.AuthenticationMiddleware"`` should be placed just after
    ``"django.contrib.auth.middleware.AuthenticationMiddleware"``:

    .. code-block:: python

        MIDDLEWARE = [
            ...,
            "django.contrib.auth.middleware.AuthenticationMiddleware",
            "sesame.middleware.AuthenticationMiddleware",
            ...,
        ]

    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        # When process_request() returns a response, return that response.
        # Otherwise continue with the next middleware or the view.
        return self.process_request(request) or self.get_response(request)

    def process_request(self, request):
        """
        Log user the in if ``request`` contains a valid token.

        Return an HTTP redirect response that removes the token from the URL
        after a successful login (except on Safari to avoid triggering ITP,
        and only when sessions are enabled).

        """
        # If django.contrib.sessions is enabled, don't update the last login,
        # because login(request, user) will do it.
        # If django.contrib.sessions is disabled, keep the default behavior of
        # updating last login only for one-time tokens.
        user = get_user(
            request,
            update_last_login=False if hasattr(request, "session") else None,
        )

        # If django.contrib.sessions is enabled and the token is valid,
        # persist the login in session.
        if hasattr(request, "session") and user is not None:
            login(request, user)
            # Once we persist the login in the session, if the authentication
            # middleware is enabled, it will set request.user in future
            # requests. We can get rid of the token in the URL by redirecting
            # to the same URL with the token removed. We only do this for GET
            # requests because redirecting POST requests doesn't work well. We
            # don't do this on Safari because it triggers the over-zealous
            # "Protection Against First Party Bounce Trackers" of ITP 2.0.
            if (
                hasattr(request, "user")
                and request.method == "GET"
                and not self.is_safari(request)
            ):
                return self.get_redirect(request)

        # If django.contrib.auth isn't enabled, set request.user.
        if not hasattr(request, "user"):
            request.user = user if user is not None else AnonymousUser()

    @staticmethod
    def is_safari(request):
        try:
            from ua_parser import user_agent_parser
        except ImportError:  # pragma: no cover
            return None
        else:
            user_agent = request.META.get("HTTP_USER_AGENT", "")
            parsed_ua = user_agent_parser.Parse(user_agent)
            return (
                parsed_ua["user_agent"]["family"] == "Safari"
                or parsed_ua["os"]["family"] == "iOS"
            )

    @staticmethod
    def get_redirect(request):
        """
        Create an HTTP redirect response that removes the token from the URL.

        """
        params = request.GET.copy()
        params.pop(settings.TOKEN_NAME)
        url = request.path
        if params:
            url += "?" + urlencode(params)
        return redirect(url)



import functools

from django.contrib.auth import login
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ImproperlyConfigured, PermissionDenied

from .utils import get_user

__all__ = ["authenticate"]


def authenticate(
    view=None,
    scope="",
    max_age=None,
    *,
    required=True,
    permanent=False,
    override=True,
):
    """
    Decorator that looks for a signed token in the URL and authenticates a user.

    If a valid token is found, the user is available as ``request.user`` in the
    view. Else, :exc:`~django.core.exceptions.PermissionDenied` is raised,
    resulting in an HTTP 403 Forbidden error.

    :obj:`authenticate` may be applied to a view directly::

        @authenticate def view(request):
            ...

    or with arguments::

        @authenticate(scope="status-page")
        def view(request):
            ...

    If ``scope`` is set, a :ref:`scoped token <Scoped tokens>` is expected.

    If ``max_age`` is set, override the :data:`SESAME_MAX_AGE` setting.

    Set ``required`` to :obj:`False` to set ``request.user`` to an
    :class:`~django.contrib.auth.models.AnonymousUser` and execute the view when
    the token is invalid, instead of raising an exception. Then, you can check
    if ``request.user.is_authenticated``.

    :obj:`authenticate` doesn't log the user in permanently. It is intended to
    provide direct access to a specific resource without exposing all other
    private resources. This makes it more acceptable to use the less secure
    authentication mechanism provided by django-sesame.

    Set ``permanent`` to :obj:`True` to call :func:`~django.contrib.auth.login`
    after a user is authenticated.

    :obj:`authenticate` doesn't care if a user is already logged in. It looks
    for a signed token anyway and overrides ``request.user``.

    Set ``override`` to :obj:`False` to skip authentication if a user is already
    logged in.

    """
    if view is None:
        return functools.partial(
            authenticate,
            scope=scope,
            max_age=max_age,
            required=required,
            permanent=permanent,
            override=override,
        )

    @functools.wraps(view)
    def wrapper(request, *args, **kwargs):
        # Skip when a user is already logged in, unless override is enabled.
        if hasattr(request, "user") and request.user.is_authenticated and not override:
            return view(request, *args, **kwargs)

        if permanent and not hasattr(request, "session"):
            raise ImproperlyConfigured(
                "authenticate(permanent=True) requires django.contrib.sessions"
            )
        # If the user will be logged in, don't update the last login, because
        # login(request, user) will do it. Else, keep the default behavior of
        # updating last login only for one-time tokens.
        user = get_user(
            request,
            update_last_login=False if permanent else None,
            scope=scope.format(*args, **kwargs),
            max_age=max_age,
        )

        request.user = user if user is not None else AnonymousUser()

        if required and user is None:
            raise PermissionDenied

        if permanent and user is not None:
            login(request, user)  # updates the last login date

        return view(request, *args, **kwargs)

    return wrapper



from urllib.parse import urlencode

from django.contrib.auth import authenticate
from django.utils import timezone

from . import settings
from .tokens import create_token

__all__ = ["get_token", "get_parameters", "get_query_string", "get_user"]


def get_token(user, scope=""):
    """
    Generate a signed token to authenticate ``user``.

    Set ``scope`` to create a :ref:`scoped token <Scoped tokens>`.

    Use this function to create a token that :func:`get_user` will accept.

    """
    return create_token(user, scope)


def get_parameters(user, scope=""):
    """
    Generate a :class:`dict` of query string parameters to authenticate ``user``.

    Set ``scope`` to create a :ref:`scoped token <Scoped tokens>`.

    Use this function to add authentication to a URL that already contains a
    query string.

    """
    return {settings.TOKEN_NAME: create_token(user, scope)}


def get_query_string(user, scope=""):
    """
    Generate a complete query string to authenticate ``user``.

    Set ``scope`` to create a :ref:`scoped token <Scoped tokens>`.

    Use this function to add authentication to a URL that doesn't contain a
    query string.

    """
    return "?" + urlencode({settings.TOKEN_NAME: create_token(user, scope)})


def get_user(request_or_sesame, scope="", max_age=None, *, update_last_login=None):
    """
    Authenticate a user based on a signed token.

    ``request_or_sesame`` may be a :class:`~django.http.HttpRequest` containing
    a token in the URL or the token itself, created with :func:`get_token`. The
    latter supports use cases outside the HTTP request lifecycle.

    If a valid token is found, return the user. Else, return :obj:`None`.

    :func:`get_user` doesn't log the user in permanently.

    If ``scope`` is set, a :ref:`scoped token <Scoped tokens>` is expected.

    If ``max_age`` is set, override the :data:`SESAME_MAX_AGE` setting.

    If single-use tokens are enabled, :func:`get_user` invalidates the token by
    updating the user's last login date.

    Set ``update_last_login`` to :obj:`True` or :obj:`False` to always or never
    update the user's last login date, regardless of whether single-use tokens
    are enabled. Typically, if you're going to log the user in with
    :func:`~django.contrib.auth.login`, set ``update_last_login`` to
    :obj:`False` to avoid updating the last login date twice.

    """
    if isinstance(request_or_sesame, str):
        request = None
        sesame = request_or_sesame
    else:
        # request is expected to be a django.http.HttpRequest
        request = request_or_sesame
        try:
            sesame = request_or_sesame.GET.get(settings.TOKEN_NAME)
        except Exception:
            raise TypeError("get_user() expects an HTTPRequest or a token")
        if sesame is None:
            return None

    # Call authenticate() to set user.backend to the value expected by login(),
    # "sesame.backends.ModelBackend" or the dotted path to a subclass.
    user = authenticate(
        request,
        sesame=sesame,
        scope=scope,
        max_age=max_age,
    )
    if user is None:
        return None

    if update_last_login is None:
        update_last_login = settings.ONE_TIME
    if update_last_login:
        user.last_login = timezone.now()
        user.save(update_fields=["last_login"])

    return user



import logging

from . import settings

logger = logging.getLogger("sesame")

__all__ = ["create_token", "parse_token"]


def create_token(user, scope=""):
    """
    Create a signed token for a user and an optional scope.

    """
    tokens = settings.TOKENS[0]
    return tokens.create_token(user, scope)


def parse_token(token, get_user, scope="", max_age=None):
    """
    Obtain a user from a signed token and an optional scope.

    """
    for tokens in settings.TOKENS:
        # We can detect the version of a token simply by inspecting it:
        # v1 tokens contain a colon; v2 tokens don't.
        if tokens.detect_token(token):
            return tokens.parse_token(token, get_user, scope, max_age)
    else:
        logger.debug("Bad token: doesn't match a supported format")
        return None



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['sesame']

package_data = \
{'': ['*']}

install_requires = \
['django>=3.2']

extras_require = \
{'ua': ['ua-parser>=0.10']}

setup_kwargs = {
    'name': 'django-sesame',
    'version': '3.1',
    'description': 'Frictionless authentication with "Magic Links" for your Django project.',
    'long_description': '.. image:: logo/horizontal.svg\n   :width: 400px\n   :alt: django-sesame\n\ndjango-sesame provides frictionless authentication with "Magic Links" for\nyour Django project.\n\nIt generates URLs containing authentication tokens such as:\nhttps://example.com/?sesame=zxST9d0XT9xgfYLvoa9e2myN\n\nThen it authenticates users based on tokens found in URLs.\n\nMore broadly, it supports a wide range of `use cases`_ for\nstateless, token-based authentication.\n\nPlease review `(in)security`_ considerations before using django-sesame.\n\n----\n\n`Documentation is available on ReadTheDocs.`__\n\n----\n\n__ https://django-sesame.readthedocs.io/en/stable/\n\nRequirements\n------------\n\ndjango-sesame is tested with:\n\n- Django 3.2 (LTS) and 4.0;\n- Python ≥ 3.6.\n\nIt requires ``django.contrib.auth``.\n\nGetting started\n---------------\n\nInstall django-sesame:\n\n.. code-block:: console\n\n    $ pip install django-sesame\n\nOpen your project settings and add ``"sesame.backends.ModelBackend"`` to the\n``AUTHENTICATION_BACKENDS`` setting. Extending the default value, this\nlooks like:\n\n.. code-block:: python\n\n    AUTHENTICATION_BACKENDS = [\n        "django.contrib.auth.backends.ModelBackend",\n        "sesame.backends.ModelBackend",\n    ]\n\nNow, your project can authenticate users based on django-sesame tokens.\n\nQuick example\n-------------\n\nConfigure ``LoginView`` in your URLconf:\n\n.. code-block:: python\n\n    from django.urls import path\n    from sesame.views import LoginView\n\n    urlpatterns = [\n        ...,\n        path("sesame/login/", LoginView.as_view(), name="sesame-login"),\n        ...,\n    ]\n\nLoad a user from the database:\n\n.. code-block:: pycon\n\n    >>> from django.contrib.auth import get_user_model\n    >>> User = get_user_model()\n    >>> user = User.objects.first()\n\nGenerate a login URL for this user:\n\n.. code-block:: pycon\n\n    >>> from sesame.utils import get_query_string\n    >>> LOGIN_URL = "https://127.0.0.1:8000/sesame/login/"\n    >>> LOGIN_URL + get_query_string(user)\n    \'https://127.0.0.1:8000/sesame/login/?sesame=zxST9d0XT9xgfYLvoa9e2myN\'\n\n(Your token will be different from this example.)\n\nMake sure that you\'re logged out. Open the login URL. You are logged in!\n\nUse cases\n---------\n\nKnown use cases for django-sesame include:\n\n1. Login by email, an attractive option on mobile where typing passwords\n   is uncomfortable. This technique is prominently deployed by Slack.\n\n   If you\'re doing this, you should define a small ``SESAME_MAX_AGE``, perhaps\n   10 minutes.\n\n2. Authenticated links. For example, you can generate a report offline\n   and, when it\'s ready, email a link to access it. Authenticated links work\n   even if the user isn\'t logged in on the device where they\'re opening it.\n\n   Likewise, you should configure an appropriate ``SESAME_MAX_AGE``,\n   probably a few days.\n\n   Since emails may be forwarded, authenticated links shouldn\'t log the user\n   in. They should only allow access to specific views.\n\n3. Sharing links, which are a variant of authenticated links. When a user shares\n   content with a guest, you may create a phantom account for the guest and\n   generate an authenticated link tied to that account or you may reuse the\n   user\'s account.\n\n   Email forwarding is also likely in this context. Make sure that sharing links\n   don\'t log the user in.\n\n4. Authentication of WebSocket connections. The web application gets a token\n   generated by the Django server and sends it over the WebSocket connection.\n   The WebSocket server authenticate the connection with the token.\n\n   Here\'s an `example with the websockets library`__.\n\n   __ https://websockets.readthedocs.io/en/stable/howto/django.html\n\n5. Non-critical private websites, for example for a family or club site,\n   where users don\'t expect to manage a personal account with a password.\n   Authorized users can bookmark personalized authenticated URLs.\n\n   Here you can rely on the default settings because that\'s the original —\n   admittedly, niche — use case for which django-sesame was built.\n\n(In)security\n------------\n\nThe major security weakness in django-sesame is a direct consequence of the\nfeature it implements: **whoever obtains an authentication token is able to\nauthenticate to your website.**\n\nURLs end up in countless insecure places: emails, referer headers, proxy logs,\nbrowser history, etc. You can\'t avoid that. At best you can mitigate it by\ncreating short-lived or single-use tokens.\n\nOtherwise, a reasonable attempt was made to provide a secure solution. Tokens\nare secured with modern cryptography. There are configurable options for token\ninvalidation.\n',
    'author': 'Aymeric Augustin',
    'author_email': 'aymeric.augustin@m4x.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/aaugustin/django-sesame',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.2',
}


setup(**setup_kwargs)
